// pager.cpp : R\[ AvP[ṼGg |Cg`܂B
//

#include "stdafx.h"
#include <algorithm>
#include <shlwapi.h>
#include <mmsystem.h>

#pragma comment(lib, "shlwapi")
#pragma comment(lib, "winmm")

using namespace std;

// O[oϐ :
const TCHAR szWindowClass[]	= _T("pager");

//================================================================================

struct Content
{
	string	upper;
	string	under;
};

typedef vector<Content>	Contents;

void LoadContents(const char* filename, Contents& results, int randomSeed)
{
	size_t sz = results.size();
	ifstream infile(filename);
	string line;
	while(getline(infile, line))
	{
		istringstream iss(line);
		Content q;
		getline(iss, q.upper, '\t');
		getline(iss, q.under, '\t');
		results.push_back(q);
	}
	if(randomSeed != 0)
	{
		srand(randomSeed);
		std::random_shuffle(results.begin() + sz, results.end());
	}
}

void ScreenToBounds(HWND hwnd, RECT* upper, RECT* under)
{
	RECT screen;
	::GetClientRect(hwnd, &screen);
	int w = screen.right - screen.left;
	int h = screen.bottom - screen.top;
	int space = __min(w, h) / 10;
	int height = (h - space * 3) * 4 / 7;
	::SetRect(upper,
		screen.left + space,
		screen.top + space,
		screen.right - space,
		screen.top + space + height);
	::SetRect(under,
		screen.left + space,
		screen.top + space * 2 + height,
		screen.right - space,
		screen.bottom - space);
}

HFONT CalcFont(HDC hdc, const string& s, const RECT& rc, HFONT hFontPrev, LPCTSTR facename)
{
	int w = rc.right - rc.left;
	int h = rc.bottom - rc.top;
	HFONT hFontOrig = (HFONT)::GetCurrentObject(hdc, OBJ_FONT);

	LOGFONT info;
	if(hFontPrev)
	{
		::GetObject(hFontPrev, sizeof(info), &info);
	}
	else
	{
		::GetObject(hFontOrig, sizeof(info), &info);
		lstrcpy(info.lfFaceName, facename);
	}
	info.lfWidth = 0;
	int upperHeight = 512;
	int underHeight = 1;

	while(true)
	{
		HFONT hFont = ::CreateFontIndirect(&info);
		SelectObject(hdc, hFont);
		SIZE sz;
		::GetTextExtentPoint32(hdc, s.c_str(), (int)s.size(), &sz);
		if(sz.cx < w && sz.cy < h)
		{
			underHeight = info.lfHeight;
			info.lfHeight = (upperHeight + info.lfHeight) / 2;
			if(abs(upperHeight - underHeight) < 2)
			{
				SelectObject(hdc, hFontOrig);
				return hFont;
			}
		}
		else
		{
			upperHeight = info.lfHeight;
			info.lfHeight = (underHeight + info.lfHeight) / 2;
		}
		::DeleteObject(hFont);
	}
}

static inline int GetHex(TCHAR c)
{
	if('0' <= c && c <= '9')
		return c - '0';
	if('A' <= c && c <= 'F')
		return 10 + c - 'A';
	if('a' <= c && c <= 'f')
		return 10 + c - 'a';
	return 0;
}

static inline int GetByte(LPCTSTR c)
{
	return (GetHex(c[0]) << 4) + GetHex(c[1]);
}

COLORREF GetPrivateProfileColor(LPCTSTR section, LPCTSTR key, COLORREF clrDefault, LPCTSTR filename)
{
	TCHAR value[32];
	if(::GetPrivateProfileString(section, key, "", value, 32, filename) != 6)
		return clrDefault;
	return RGB(GetByte(value), GetByte(value+2), GetByte(value+4));
}

//================================================================================

Contents	theContents;
size_t		theCurrent;
RECT		theUpperBounds;
RECT		theUnderBounds;
HFONT		theUpperFont;
HFONT		theUnderFont;
HFONT		theNumberFont;
TCHAR		theBasePath[MAX_PATH];
TCHAR		theProfilePath[MAX_PATH];
TCHAR		theSoundOK[MAX_PATH];
TCHAR		theSoundNG[MAX_PATH];
TCHAR		theUpperFontName[32];
TCHAR		theUnderFontName[32];
COLORREF	theBackgroundColor;
COLORREF	theUpperFontColor;
COLORREF	theUnderFontColor;
COLORREF	theNumberFontColor;
const size_t	JUMPLEN = 32;
TCHAR		theJump[JUMPLEN];

//================================================================================

// ̃R[h W[Ɋ܂܂֐̐錾]܂ :
BOOL				InitInstance(HINSTANCE, int);
LRESULT CALLBACK	WndProc(HWND, UINT, WPARAM, LPARAM);

static void InitContents(int randomSeed)
{
	TCHAR path[MAX_PATH];
	PathCombine(path, theBasePath, "contents.txt");

	TCHAR upper[MAX_PATH], under[MAX_PATH];
	// First
	if(GetPrivateProfileString("Text", "FirstUpper", "", upper, MAX_PATH, theProfilePath) > 0 &&
		GetPrivateProfileString("Text", "FirstUnder", "", under, MAX_PATH, theProfilePath) > 0)
	{
		Content q;
		q.upper = upper;
		q.under = under;
		theContents.push_back(q);
	}

	LoadContents(path, theContents, randomSeed);

	// Last
	if(GetPrivateProfileString("Text", "LastUpper", "", upper, MAX_PATH, theProfilePath) > 0 &&
		GetPrivateProfileString("Text", "LastUnder", "", under, MAX_PATH, theProfilePath) > 0)
	{
		Content q;
		q.upper = upper;
		q.under = under;
		theContents.push_back(q);
	}
}

int APIENTRY _tWinMain(HINSTANCE hInstance, HINSTANCE, LPTSTR lpCmdLine, int nCmdShow)
{
	::GetModuleFileName(NULL, theBasePath, MAX_PATH);
	::PathRemoveFileSpec(theBasePath);

	PathCombine(theProfilePath, theBasePath, "pager.ini");
	PathCombine(theSoundOK, theBasePath, "sound\\OK.wav");
	PathCombine(theSoundNG, theBasePath, "sound\\NG.wav");

	GetPrivateProfileString("Look", "UpperFontName", "HGۺ޼M-PRO", theUpperFontName, 32, theProfilePath);
	GetPrivateProfileString("Look", "UnderFontName", "Tahoma", theUpperFontName, 32, theProfilePath);

	theBackgroundColor = GetPrivateProfileColor("Look", "BackgroundColor", RGB(224, 224, 255), theProfilePath);
	theUpperFontColor  = GetPrivateProfileColor("Look", "UpperFontColor", RGB(128, 0, 0), theProfilePath);
	theUnderFontColor  = GetPrivateProfileColor("Look", "UnderFontColor", RGB(0, 0, 128), theProfilePath);

	theNumberFontColor = GetPrivateProfileColor("Look", "NumberFontColor", RGB(0, 0, 0), theProfilePath);
	LOGFONT numberFont;
	::GetObject(::GetStockObject(DEFAULT_GUI_FONT), sizeof(numberFont), &numberFont);
	numberFont.lfHeight = GetPrivateProfileInt("Look", "NumberFontHeight", 0, theProfilePath);
	numberFont.lfWidth = 0;
	TCHAR numberFontName[32];
	if(GetPrivateProfileString("Look", "NumberFontName", "Tahoma", numberFontName, 32, theProfilePath) > 0)
		lstrcpy(numberFont.lfFaceName, numberFontName);
	theNumberFont = ::CreateFontIndirect(&numberFont);

	int randomSeed = GetPrivateProfileInt("Text", "Random", 0, theProfilePath);
	InitContents(randomSeed);

	if (!InitInstance(hInstance, nCmdShow)) 
	{
		return FALSE;
	}

	MSG msg;
	while (GetMessage(&msg, NULL, 0, 0)) 
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return (int) msg.wParam;
}

BOOL InitInstance(HINSTANCE hInstance, int nCmdShow)
{
	WNDCLASSEX wcex = { sizeof(WNDCLASSEX) };
	wcex.style			= CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc	= (WNDPROC)WndProc;
	wcex.hInstance		= hInstance;
	wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground	= ::CreateSolidBrush(theBackgroundColor);
	wcex.lpszClassName	= szWindowClass;
	if(!RegisterClassEx(&wcex))
		return FALSE;

	RECT rcDesktop;
	::GetWindowRect(::GetDesktopWindow(), &rcDesktop);

	HWND hwnd = CreateWindow(szWindowClass, NULL, WS_POPUP,
		rcDesktop.left, rcDesktop.top, rcDesktop.right-rcDesktop.left, rcDesktop.bottom-rcDesktop.top,
		NULL, NULL, hInstance, NULL);

	if (!hwnd)
		return FALSE;

	ShowWindow(hwnd, nCmdShow);
	return TRUE;
}

static void OnResetFont(HWND hwnd)
{
	HDC hdc = ::GetDC(hwnd);
	if(HFONT hFont = CalcFont(hdc, theContents[theCurrent].upper, theUpperBounds, theUpperFont, theUpperFontName))
	{
		::DeleteObject(theUpperFont);
		theUpperFont = hFont;
	}
	if(HFONT hFont = CalcFont(hdc, theContents[theCurrent].under, theUnderBounds, theUnderFont, theUnderFontName))
	{
		::DeleteObject(theUnderFont);
		theUnderFont = hFont;
	}
	::ReleaseDC(hwnd, hdc);
}

static void OnPaint(HWND hwnd, HDC hdc)
{
	const Content& q = theContents[theCurrent];
	SetBkMode(hdc, TRANSPARENT);

	HGDIOBJ hFontOrig = SelectObject(hdc, theUpperFont);
	SetTextColor(hdc, theUpperFontColor);
	DrawText(hdc, q.upper.c_str(), -1, &theUpperBounds, DT_CENTER | DT_VCENTER | DT_SINGLELINE);

	SetTextColor(hdc, theUnderFontColor);
	SelectObject(hdc, theUnderFont);
	DrawText(hdc, q.under.c_str(), -1, &theUnderBounds, DT_CENTER | DT_VCENTER | DT_SINGLELINE);

	TCHAR buffer[128];
	wsprintf(buffer, "%d/%d", theCurrent+1, theContents.size());
	RECT rcNumber;
	::GetWindowRect(hwnd, &rcNumber);
	rcNumber.top = theUnderBounds.bottom;
	rcNumber.bottom -= 10;
	rcNumber.right  -= 10;
	SetTextColor(hdc, theNumberFontColor);
	SelectObject(hdc, theNumberFont);
	DrawText(hdc, buffer, -1, &rcNumber, DT_RIGHT | DT_BOTTOM | DT_SINGLELINE);

	SelectObject(hdc, hFontOrig);
}

static void JumpToPage(HWND hwnd, size_t page)
{
	theCurrent = page;
	memset(theJump, 0, sizeof(theJump));
	OnResetFont(hwnd);
	::InvalidateRect(hwnd, NULL, TRUE);
}

static bool NextPage(HWND hwnd)
{
	if(theCurrent + 1 >= theContents.size())
		return false;
	JumpToPage(hwnd, theCurrent + 1);
	return true;
}

static bool PrevPage(HWND hwnd)
{
	if(theCurrent < 1)
		return false;
	JumpToPage(hwnd, theCurrent - 1);
	return true;
}

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	PAINTSTRUCT ps;
	HDC hdc;

	switch (message) 
	{
	case WM_PAINT:
		hdc = BeginPaint(hwnd, &ps);
		OnPaint(hwnd, hdc);
		EndPaint(hwnd, &ps);
		break;
	case WM_SIZE:
		ScreenToBounds(hwnd, &theUpperBounds, &theUnderBounds);
		OnResetFont(hwnd);
		break;
	case WM_DESTROY:
		PostQuitMessage(0);
		break;
	case WM_KEYDOWN:
		switch(wParam)
		{
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
		{
			size_t len = lstrlen(theJump);
			if(len < JUMPLEN)
				theJump[len] = (TCHAR)wParam;
			return 0;
		}
		case VK_RETURN:
			if(lstrlen(theJump) > 0)
			{
				size_t jump = (size_t)atoi(theJump) - 1;
				if(jump < theContents.size())
					JumpToPage(hwnd, jump);
				else
					memset(theJump, 0, sizeof(theJump));
			}
			return 0;
		case VK_BACK:
			PrevPage(hwnd);
			return 0;
		case VK_ESCAPE:
			::PostMessage(hwnd, WM_CLOSE, 0, 0);
			return 0;
		case VK_LEFT:
			PlaySound(theSoundOK, NULL, SND_ASYNC | SND_FILENAME);
			return 0;
		case VK_RIGHT:
			PlaySound(theSoundNG, NULL, SND_ASYNC | SND_FILENAME);
			return 0;
		default:
			NextPage(hwnd);
			break;
		}
		break;
	case WM_LBUTTONDOWN:
		PlaySound(theSoundOK, NULL, SND_ASYNC | SND_FILENAME);
		break;
	case WM_RBUTTONDOWN:
		PlaySound(theSoundNG, NULL, SND_ASYNC | SND_FILENAME);
		break;
	case WM_MOUSEWHEEL:
		if(GET_WHEEL_DELTA_WPARAM(wParam) < 0)
			NextPage(hwnd);
		else
			PrevPage(hwnd);
		break;
	default:
		return DefWindowProc(hwnd, message, wParam, lParam);
	}
	return 0;
}
